syntax = "proto3";

package text_data;

message Semantics {
    repeated uint32 values = 1;
}

message Sentence {
    repeated string texts = 1;
    repeated Semantics semantics = 3;
}

message TextData {
    string source = 1;
    string name = 2;
    repeated Sentence sentences = 4;
}

message SampledData {
    string source = 1;
    string name = 2;
    repeated Sentence samples = 3;
}
